// Copyright 2024, 2025 New Vector Ltd.
// Copyright 2023, 2024 The Matrix.org Foundation C.I.C.
//
// SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
// Please see LICENSE files in the repository root for full details.

mod error;
mod legacy;
mod modern;

pub use self::{legacy::SynapseConnection as LegacySynapseConnection, modern::SynapseConnection};
